@@ -1,13 +1,35 @@
 import os
 from abc import abstractmethod
 from enum import Enum
-from typing import List, Optional, Union
-
-from ordered_set import OrderedSet
+from typing import List, Optional, Union, Iterable
 
 ELEMENT_ATTRIBUTE_PREFIX = "}ELEMENTATTRIBUTES_"
 
 
+class DescFlag(Enum):
+    SELF = 1
+    AFTER = 2
+    BEFORE = 3
+    BEFORE_AND_AFTER = 4
+    SELF_AND_AFTER = 5
+    SELF_AND_BEFORE = 6
+    SELF_BEFORE_AFTER = 7
+    LEAVES = 8
+
+    def __str__(self):
+        return self.name
+
+    @classmethod
+    def _missing_(cls, value: str):
+        if value is None:
+            return None
+        for member in cls:
+            if member.name.lower() == value.replace(" ", "").lower():
+                return member
+        # default
+        raise ValueError(f"Invalid Desc Flag type: '{value}'")
+
+
 class Order(Enum):
     ASC = 1
     DESC = 2
@@ -26,11 +48,30 @@ class Order(Enum):
         raise ValueError(f"Invalid order type: '{value}'")
 
 
+class ElementType(Enum):
+    NUMERIC = 1
+    STRING = 2
+    CONSOLIDATED = 3
+
+    def __str__(self):
+        return self.name
+
+    @classmethod
+    def _missing_(cls, value: str):
+        for member in cls:
+            if member.name.lower() == value.replace(" ", "").lower():
+                return member
+        # default
+        raise ValueError(f"Invalid element type: '{value}'")
+
+
 def normalize(name: str) -> str:
-    return name.upper().replace(" ", "").replace("]", "]]")
+    return name.lower().replace(" ", "").replace("]", "]]")
 
 
 class Member:
+    # control if full element unique name is used for members without explicit hierarchy
+    SHORT_NOTATION = False
 
     def __init__(self, dimension: str, hierarchy: str, element: str):
         self.dimension = dimension
@@ -40,6 +81,8 @@ class Member:
 
     @classmethod
     def build_unique_name(cls, dimension, hierarchy, element) -> str:
+        if cls.SHORT_NOTATION and dimension == hierarchy:
+            return f"[{normalize(dimension)}].[{normalize(element)}]"
         return f"[{normalize(dimension)}].[{normalize(hierarchy)}].[{normalize(element)}]"
 
     @staticmethod
@@ -95,34 +138,87 @@ class CalculatedMember(Member):
     @staticmethod
     def avg(dimension: str, hierarchy: str, element: str, cube: str, mdx_set: 'MdxHierarchySet',
             mdx_tuple: 'MdxTuple'):
-        calculation = f"AVG({mdx_set.to_mdx()},[{cube.upper()}].{mdx_tuple.to_mdx()})"
+        calculation = f"AVG({mdx_set.to_mdx()},[{cube.lower()}].{mdx_tuple.to_mdx()})"
         return CalculatedMember(dimension, hierarchy, element, calculation)
 
     @staticmethod
     def sum(dimension: str, hierarchy: str, element: str, cube: str, mdx_set: 'MdxHierarchySet',
             mdx_tuple: 'MdxTuple'):
-        calculation = f"SUM({mdx_set.to_mdx()},[{cube.upper()}].{mdx_tuple.to_mdx()})"
+        calculation = f"SUM({mdx_set.to_mdx()},[{cube.lower()}].{mdx_tuple.to_mdx()})"
         return CalculatedMember(dimension, hierarchy, element, calculation)
 
     @staticmethod
     def lookup(dimension: str, hierarchy: str, element: str, cube: str, mdx_tuple: 'MdxTuple'):
-        calculation = f"[{cube.upper()}].{mdx_tuple.to_mdx()}"
+        calculation = f"[{cube.lower()}].{mdx_tuple.to_mdx()}"
         return CalculatedMember(dimension, hierarchy, element, calculation)
 
     @staticmethod
     def lookup_attribute(dimension: str, hierarchy: str, element: str, attribute_dimension: str, attribute: str):
-        attribute_cube = ELEMENT_ATTRIBUTE_PREFIX + attribute_dimension.upper()
-        calculation = f"[{attribute_cube}].([{attribute_cube}].[{attribute.upper()}])"
+        attribute_cube = ELEMENT_ATTRIBUTE_PREFIX + attribute_dimension.lower()
+        calculation = f"[{attribute_cube}].([{attribute_cube}].[{attribute.lower()}])"
         return CalculatedMember(dimension, hierarchy, element, calculation)
 
     def to_mdx(self):
         return f"MEMBER {self.unique_name} AS {self.calculation}"
 
 
+class MdxLevelExpression:
+
+    def __init__(self, dimension: str, hierarchy: Optional[str] = None):
+        self.dimension = normalize(dimension)
+        self.hierarchy = normalize(hierarchy) if hierarchy else self.dimension
+
+    @abstractmethod
+    def to_mdx(self) -> str:
+        pass
+
+    @staticmethod
+    def level_number(level: int, dimension: str, hierarchy: str = None) -> 'MdxLevelExpression':
+        return LevelNumberExpression(level, dimension, hierarchy)
+
+    @staticmethod
+    def level_name(level: str, dimension: str, hierarchy: str = None) -> 'MdxLevelExpression':
+        return LevelNameExpression(level, dimension, hierarchy)
+
+    @staticmethod
+    def member_level(member: Union[str, Member]) -> 'MdxLevelExpression':
+        return MemberLevelExpression(member)
+
+
+class LevelNumberExpression(MdxLevelExpression):
+
+    def __init__(self, level_number: int, dimension: str, hierarchy: str = None):
+        super(LevelNumberExpression, self).__init__(dimension, hierarchy)
+        self.level = level_number
+
+    def to_mdx(self) -> str:
+        return f"[{self.dimension}].[{self.hierarchy}].LEVELS({self.level})"
+
+
+class LevelNameExpression(MdxLevelExpression):
+
+    def __init__(self, level_name: str, dimension: str, hierarchy: str = None):
+        super(LevelNameExpression, self).__init__(dimension, hierarchy)
+        self.level = level_name
+
+    def to_mdx(self) -> str:
+        return f"[{self.dimension}].[{self.hierarchy}].LEVELS(\'{self.level}\')"
+
+
+class MemberLevelExpression(MdxLevelExpression):
+
+    def __init__(self, member: Member):
+        super(MemberLevelExpression, self).__init__(member.dimension, member.hierarchy)
+        self.member = member
+
+    def to_mdx(self) -> str:
+        return f"{self.member.unique_name}.LEVEL"
+
+
 class MdxTuple:
 
     def __init__(self, members):
-        self.members = OrderedSet(members)
+        self.members = list(members)
 
     @staticmethod
     def of(*args: Union[str, Member]) -> 'MdxTuple':
@@ -140,10 +236,10 @@ class MdxTuple:
     def add_member(self, member: Union[str, Member]):
         if isinstance(member, str):
             member = Member.of(member)
-        self.members.add(member)
+        self.members.append(member)
 
     def is_empty(self) -> bool:
-        return len(self.members) == 0
+        return not self.members
 
     def to_mdx(self) -> str:
         return f"({','.join(member.unique_name for member in self.members)})"
@@ -152,22 +248,75 @@ class MdxTuple:
         return len(self.members)
 
 
-class MdxHierarchySet:
-
-    def __init__(self, dimension: str, hierarchy: Optional[str] = None):
-        self.dimension = normalize(dimension)
-        self.hierarchy = normalize(hierarchy) if hierarchy else self.dimension
+class MdxSet:
 
     @abstractmethod
     def to_mdx(self) -> str:
         pass
 
+    @staticmethod
+    def cross_joins(sets: List['MdxSet']) -> 'MdxSet':
+        return CrossJoinMdxSet(sets)
+
+    @staticmethod
+    def unions(sets: List['MdxSet'], allow_duplicates: bool = False) -> 'MdxSet':
+        return MultiUnionSet(sets, allow_duplicates)
+
+    @staticmethod
+    def tuples(tuples: Iterable['MdxTuple']) -> 'MdxSet':
+        return TuplesSet(tuples)
+
+
+class CrossJoinMdxSet(MdxSet):
+    def __init__(self, sets: List['MdxSet']):
+        if not sets:
+            raise RuntimeError('sets must not be empty')
+        self.sets = sets
+
+    def to_mdx(self) -> str:
+        return f"{{{' * '.join(set_.to_mdx() for set_ in self.sets)}}}"
+
+
+class TuplesSet(MdxSet):
+    def __init__(self, tuples: Iterable[MdxTuple]):
+        self.tuples = tuples
+
+    def to_mdx(self) -> str:
+        return f"{{ {','.join(tupl.to_mdx() for tupl in self.tuples)} }}"
+
+
+class MultiUnionSet(MdxSet):
+
+    def __init__(self, sets: List[MdxSet], allow_duplicates: bool = False):
+        if not sets:
+            raise RuntimeError('sets must not be empty')
+
+        self.sets = sets
+        self.allow_duplicates = allow_duplicates
+
+    def to_mdx(self) -> str:
+        if self.allow_duplicates:
+            return f"{{{','.join(set_.to_mdx() for set_ in self.sets)}}}"
+        else:
+            return f"{{{' + '.join(set_.to_mdx() for set_ in self.sets)}}}"
+
+
+class MdxHierarchySet(MdxSet):
+
+    def __init__(self, dimension: str, hierarchy: Optional[str] = None):
+        self.dimension = normalize(dimension)
+        self.hierarchy = normalize(hierarchy) if hierarchy else self.dimension
+
     def to_clipboard(self):
         mdx = self.to_mdx()
         command = 'echo | set /p nul="' + mdx + '"| clip'
         os.system(command)
         print(mdx)
 
+    @abstractmethod
+    def to_mdx(self) -> str:
+        pass
+
     @staticmethod
     def tm1_subset_all(dimension: str, hierarchy: str = None) -> 'MdxHierarchySet':
         return Tm1SubsetAllHierarchySet(dimension, hierarchy)
@@ -180,6 +329,10 @@ class MdxHierarchySet:
     def tm1_subset_to_set(dimension: str, hierarchy: str, subset: str) -> 'MdxHierarchySet':
         return Tm1SubsetToSetHierarchySet(dimension, hierarchy, subset)
 
+    @staticmethod
+    def tm1_dimension_subset_to_set(dimension: str, subset: str) -> 'MdxHierarchySet':
+        return Tm1SubsetToSetHierarchySet(dimension, dimension, subset)
+
     @staticmethod
     def all_consolidations(dimension: str, hierarchy: str = None) -> 'MdxHierarchySet':
         return AllCElementsHierarchySet(dimension, hierarchy)
@@ -249,15 +402,24 @@ class MdxHierarchySet:
         return DrillDownLevelHierarchySet(member)
 
     @staticmethod
-    def descendants(member: Union[str, Member]) -> 'MdxHierarchySet':
+    def descendants(member: Union[str, Member], level_or_depth: Union[MdxLevelExpression, int] = None,
+                    desc_flag: Union[str, Order] = None) -> 'MdxHierarchySet':
         if isinstance(member, str):
             member = Member.of(member)
-        return DescendantsHierarchySet(member)
+        return DescendantsHierarchySet(member, level_or_depth, desc_flag)
 
     @staticmethod
     def from_str(dimension: str, hierarchy: str, mdx: str):
         return StrHierarchySet(dimension, hierarchy, mdx)
 
+    @staticmethod
+    def range(start_member: Union[str, Member], end_member: Union[str, Member]) -> 'MdxHierarchySet':
+        if isinstance(start_member, str):
+            start_member = Member.of(start_member)
+        if isinstance(end_member, str):
+            end_member = Member.of(end_member)
+        return RangeHierarchySet(start_member, end_member)
+
     def filter_by_attribute(self, attribute_name: str, attribute_values: List,
                             operator: Optional[str] = '=') -> 'MdxHierarchySet':
         return FilterByAttributeHierarchySet(self, attribute_name, attribute_values, operator)
@@ -268,6 +430,9 @@ class MdxHierarchySet:
     def filter_by_level(self, level: int) -> 'MdxHierarchySet':
         return Tm1FilterByLevelHierarchySet(self, level)
 
+    def filter_by_element_type(self, element_type: Union[ElementType, str]) -> 'MdxHierarchySet':
+        return Tm1FilterByElementTypeHierarchySet(self, element_type)
+
     def filter_by_cell_value(self, cube: str, mdx_tuple: MdxTuple, operator: str, value) -> 'MdxHierarchySet':
         return FilterByCellValueHierarchySet(self, cube, mdx_tuple, operator, value)
 
@@ -293,8 +458,8 @@ class MdxHierarchySet:
     def bottom_count(self, cube, mdx_tuple, top) -> 'MdxHierarchySet':
         return BottomCountHierarchySet(self, cube, mdx_tuple, top)
 
-    def union(self, other_set: 'MdxHierarchySet') -> 'MdxHierarchySet':
-        return UnionHierarchySet(self, other_set)
+    def union(self, other_set: 'MdxHierarchySet', allow_duplicates: bool = False) -> 'MdxHierarchySet':
+        return UnionHierarchySet(self, other_set, allow_duplicates)
 
     def intersect(self, other_set: 'MdxHierarchySet') -> 'MdxHierarchySet':
         return IntersectHierarchySet(self, other_set)
@@ -312,9 +477,9 @@ class MdxHierarchySet:
     def generate_attribute_to_member(self, attribute: str, dimension: str, hierarchy: str = None):
         return GenerateAttributeToMemberSet(self, attribute, dimension, hierarchy)
 
-    def tm1_drill_down_member(self, all: bool = True, other_set: 'MdxHierarchySet' = None,
+    def tm1_drill_down_member(self, other_set: 'MdxHierarchySet' = None,
                               recursive: bool = True) -> 'MdxHierarchySet':
-        return Tm1DrillDownMemberSet(self, all, other_set, recursive)
+        return Tm1DrillDownMemberSet(self, other_set, recursive)
 
 
 class Tm1SubsetAllHierarchySet(MdxHierarchySet):
@@ -439,13 +604,15 @@ class ChildrenHierarchySet(MdxHierarchySet):
 
 class Tm1DrillDownMemberSet(MdxHierarchySet):
 
-    def __init__(self, underlying_hierarchy_set: MdxHierarchySet, all: bool = True, other_set: 'MdxHierarchySet' = None,
+    def __init__(self, underlying_hierarchy_set: MdxHierarchySet, other_set: 'MdxHierarchySet' = None,
                  recursive: bool = True):
-        super(Tm1DrillDownMemberSet, self).__init__(underlying_hierarchy_set.dimension,
-                                                    underlying_hierarchy_set.hierarchy)
+        super(Tm1DrillDownMemberSet, self).__init__(
+            underlying_hierarchy_set.dimension,
+            underlying_hierarchy_set.hierarchy)
         self.underlying_hierarchy_set = underlying_hierarchy_set
+
         if other_set:
-            self.set2 = other_set
+            self.set2 = other_set.to_mdx()
         else:
             self.set2 = "ALL"
 
@@ -455,10 +622,7 @@ class Tm1DrillDownMemberSet(MdxHierarchySet):
             self.recursive = ""
 
     def to_mdx(self) -> str:
-        if self.set2 == "ALL":
-            return f"{{TM1DRILLDOWNMEMBER({self.underlying_hierarchy_set.to_mdx()}, {self.set2}{self.recursive})}}"
-        else:
-            return f"{{TM1DRILLDOWNMEMBER({self.underlying_hierarchy_set.to_mdx()}, {self.set2.to_mdx()}{self.recursive})}}"
+        return f"{{TM1DRILLDOWNMEMBER({self.underlying_hierarchy_set.to_mdx()}, {self.set2}{self.recursive})}}"
 
 
 class DrillDownLevelHierarchySet(MdxHierarchySet):
@@ -473,12 +637,31 @@ class DrillDownLevelHierarchySet(MdxHierarchySet):
 
 class DescendantsHierarchySet(MdxHierarchySet):
 
-    def __init__(self, member: Member):
+    def __init__(self, member: Member, level_or_depth: Union[int, MdxLevelExpression] = None,
+                 description_flag: DescFlag = None):
         super(DescendantsHierarchySet, self).__init__(member.dimension, member.hierarchy)
         self.member = member
+        self.level_or_depth = level_or_depth
+        self.descFlag = DescFlag(description_flag) if description_flag is not None else None
 
     def to_mdx(self) -> str:
-        return f"{{DESCENDANTS({self.member.unique_name})}}"
+        if isinstance(self.level_or_depth, MdxLevelExpression):
+            level_expression = f', {self.level_or_depth.to_mdx()}'
+        else:
+            level_expression = f",{self.level_or_depth}" if self.level_or_depth is not None else ''
+
+        flag = f", {self.descFlag}" if self.descFlag is not None else ''
+        return f"{{DESCENDANTS({self.member.unique_name}{level_expression}{flag})}}"
+
+
+class RangeHierarchySet(MdxHierarchySet):
+    def __init__(self, start_member: Member, end_member: Member):
+        super(RangeHierarchySet, self).__init__(start_member.dimension, start_member.hierarchy)
+        self._start_member = start_member
+        self._end_member = end_member
+
+    def to_mdx(self) -> str:
+        return f"{{{self._start_member.unique_name}:{self._end_member.unique_name}}}"
 
 
 class Tm1SubsetToSetHierarchySet(MdxHierarchySet):
@@ -549,6 +732,19 @@ class Tm1FilterByLevelHierarchySet(MdxHierarchySet):
         return f"{{TM1FILTERBYLEVEL({self.underlying_hierarchy_set.to_mdx()},{self.level})}}"
 
 
+class Tm1FilterByElementTypeHierarchySet(MdxHierarchySet):
+
+    def __init__(self, underlying_hierarchy_set: MdxHierarchySet, element_type: Union[ElementType, str]):
+        super(Tm1FilterByElementTypeHierarchySet, self).__init__(underlying_hierarchy_set.dimension,
+                                                                 underlying_hierarchy_set.hierarchy)
+        self.underlying_hierarchy_set = underlying_hierarchy_set
+        self.element_type = ElementType(element_type)
+
+    def to_mdx(self) -> str:
+        return f"{{FILTER({self.underlying_hierarchy_set.to_mdx()},[{self.dimension}].[{self.hierarchy}]" \
+               f".CURRENTMEMBER.PROPERTIES('ELEMENT_TYPE')='{self.element_type.value}')}}"
+
+
 class FilterByCellValueHierarchySet(MdxHierarchySet):
 
     def __init__(self, underlying_hierarchy_set: MdxHierarchySet, cube: str, mdx_tuple: MdxTuple, operator, value):
@@ -576,13 +772,13 @@ class FilterByInstr(MdxHierarchySet):
         self.underlying_hierarchy_set = underlying_hierarchy_set
         self.cube = normalize(cube)
         self.mdx_tuple = mdx_tuple
-        self.substring = substring.upper() if case_insensitive else substring
+        self.substring = substring.lower() if case_insensitive else substring
         self.operator = operator
         self.position = position
         self.case_insensitive = case_insensitive
 
     def to_mdx(self) -> str:
-        return f"{{FILTER({self.underlying_hierarchy_set.to_mdx()},INSTR({'UCASE(' if self.case_insensitive else ''}" \
+        return f"{{FILTER({self.underlying_hierarchy_set.to_mdx()},INSTR({'LCASE(' if self.case_insensitive else ''}" \
                f"[{self.cube}].{self.mdx_tuple.to_mdx()}{')' if self.case_insensitive else ''},'{self.substring}')" \
                f"{self.operator}{self.position})}}"
 
@@ -664,13 +860,15 @@ class SubsetHierarchySet(MdxHierarchySet):
 
 class UnionHierarchySet(MdxHierarchySet):
 
-    def __init__(self, underlying_hierarchy_set: MdxHierarchySet, other_hierarchy_set: MdxHierarchySet):
+    def __init__(self, underlying_hierarchy_set: MdxHierarchySet, other_hierarchy_set: MdxHierarchySet,
+                 allow_duplicates: bool):
         super(UnionHierarchySet, self).__init__(underlying_hierarchy_set.dimension, underlying_hierarchy_set.hierarchy)
         self.underlying_hierarchy_set = underlying_hierarchy_set
         self.other_hierarchy_set = other_hierarchy_set
+        self.allow_duplicates = allow_duplicates
 
     def to_mdx(self) -> str:
-        return f"{{UNION({self.underlying_hierarchy_set.to_mdx()},{self.other_hierarchy_set.to_mdx()})}}"
+        return f"{{UNION({self.underlying_hierarchy_set.to_mdx()},{self.other_hierarchy_set.to_mdx()}{', ALL' if self.allow_duplicates else ''})}}"
 
 
 class IntersectHierarchySet(MdxHierarchySet):
@@ -733,8 +931,8 @@ class GenerateAttributeToMemberSet(MdxHierarchySet):
             underlying_hierarchy_set.dimension,
             underlying_hierarchy_set.hierarchy)
         self.underlying_hierarchy_set = underlying_hierarchy_set
-        self.dimension = dimension.upper()
-        self.hierarchy = hierarchy.upper() if hierarchy else self.dimension
+        self.dimension = dimension.lower()
+        self.hierarchy = hierarchy.lower() if hierarchy else self.dimension
         self.attribute = attribute
 
     def to_mdx(self) -> str:
@@ -744,8 +942,8 @@ class GenerateAttributeToMemberSet(MdxHierarchySet):
 
 class MdxAxis:
     def __init__(self):
-        self.tuples = list()
-        self.dim_sets = list()
+        self.tuples: List[MdxTuple] = list()
+        self.dim_sets: List[MdxSet] = list()
         self.non_empty = False
 
     @staticmethod
@@ -758,11 +956,14 @@ class MdxAxis:
 
         self.tuples.append(mdx_tuple)
 
-    def add_dim_set(self, mdx_hierarchy_set: MdxHierarchySet):
+    def add_set(self, mdx_set: MdxSet):
         if bool(self.tuples):
             raise ValueError("Can not add set to axis that contains tuples")
 
-        self.dim_sets.append(mdx_hierarchy_set)
+        if not isinstance(mdx_set, MdxSet):
+            raise ValueError("Can not add MDX Tuples to axis using set method")
+
+        self.dim_sets.append(mdx_set)
 
     def is_empty(self) -> bool:
         return not self.dim_sets and not self.tuples
@@ -816,20 +1017,23 @@ class MdxBuilder:
         self._tm1_ignore_bad_tuples = ignore
         return self
 
-    def _add_tuple_to_axis(self, axis: MdxAxis, *args: Union[str, Member]) -> 'MdxBuilder':
-        mdx_tuple = MdxTuple.of(*args)
+    def _add_tuple_to_axis(self, axis: MdxAxis, *args: Union[str, Member, MdxTuple]) -> 'MdxBuilder':
+        if isinstance(args[0], MdxTuple):
+            mdx_tuple = args[0]
+        else:
+            mdx_tuple = MdxTuple.of(*args)
         axis.add_tuple(mdx_tuple)
         return self
 
-    def add_member_tuple_to_axis(self, axis: int, *args: Union[str, Member]) -> 'MdxBuilder':
+    def add_member_tuple_to_axis(self, axis: int, *args: Union[str, Member, MdxTuple]) -> 'MdxBuilder':
         if axis not in self.axes:
             self.axes[axis] = MdxAxis.empty()
         return self._add_tuple_to_axis(self.axes[axis], *args)
 
-    def add_member_tuple_to_columns(self, *args: Union[str, Member]) -> 'MdxBuilder':
+    def add_member_tuple_to_columns(self, *args: Union[str, Member, MdxTuple]) -> 'MdxBuilder':
         return self.add_member_tuple_to_axis(0, *args)
 
-    def add_member_tuple_to_rows(self, *args: Union[str, Member]) -> 'MdxBuilder':
+    def add_member_tuple_to_rows(self, *args: Union[str, Member, MdxTuple]) -> 'MdxBuilder':
         return self.add_member_tuple_to_axis(1, *args)
 
     def add_hierarchy_set_to_row_axis(self, mdx_hierarchy_set: MdxHierarchySet) -> 'MdxBuilder':
@@ -838,13 +1042,29 @@ class MdxBuilder:
     def add_hierarchy_set_to_column_axis(self, mdx_hierarchy_set: MdxHierarchySet) -> 'MdxBuilder':
         return self.add_hierarchy_set_to_axis(0, mdx_hierarchy_set)
 
+    def add_set_to_row_axis(self, mdx_set: MdxSet) -> 'MdxBuilder':
+        return self.add_set_to_axis(1, mdx_set)
+
+    def add_set_to_column_axis(self, mdx_set: MdxSet) -> 'MdxBuilder':
+        return self.add_set_to_axis(0, mdx_set)
+
     def add_hierarchy_set_to_axis(self, axis: int, mdx_hierarchy_set: MdxHierarchySet) -> 'MdxBuilder':
+        return self.add_set_to_axis(axis, mdx_hierarchy_set)
+
+    def add_set_to_axis(self, axis: int, mdx_set: MdxSet) -> 'MdxBuilder':
         if axis not in self.axes:
             self.axes[axis] = MdxAxis.empty()
 
-        self.axes[axis].add_dim_set(mdx_hierarchy_set)
+        self.axes[axis].add_set(mdx_set)
         return self
 
+    def add_empty_set_to_axis(self, axis: int):
+        if axis in self.axes:
+            raise ValueError(f"axis: '{axis}' must be empty")
+
+        hierarchy_set = MdxHierarchySet.from_str("", "", "{}")
+        return self.add_hierarchy_set_to_axis(axis, hierarchy_set)
+
     def add_member_to_where(self, member: Union[str, Member]) -> 'MdxBuilder':
         self._where.add_member(member)
         return self
